from dataclasses import dataclass, field


@dataclass
class Connection:
    """
    Connection.

    :ivar device_id_ref_0: A, (required)
    :ivar device_element_id_ref_0: B, (required)
    :ivar device_id_ref_1: C, (required)
    :ivar device_element_id_ref_1: D, (required)
    """

    class Meta:
        name = "CNN"

    device_id_ref_0: None | str = field(
        default=None,
        metadata={
            "name": "A",
            "full_name": "DeviceIdRef_0",
            "type": "Attribute",
            "required": True,
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(DVC|DVC-)([0-9])+",
        },
    )
    device_element_id_ref_0: None | str = field(
        default=None,
        metadata={
            "name": "B",
            "full_name": "DeviceElementIdRef_0",
            "type": "Attribute",
            "required": True,
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(DET|DET-)([0-9])+",
        },
    )
    device_id_ref_1: None | str = field(
        default=None,
        metadata={
            "name": "C",
            "full_name": "DeviceIdRef_1",
            "type": "Attribute",
            "required": True,
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(DVC|DVC-)([0-9])+",
        },
    )
    device_element_id_ref_1: None | str = field(
        default=None,
        metadata={
            "name": "D",
            "full_name": "DeviceElementIdRef_1",
            "type": "Attribute",
            "required": True,
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(DET|DET-)([0-9])+",
        },
    )



from dataclasses import dataclass, field


@dataclass
class Customer:
    """
    Customer.

    :ivar id: A, (required)
    :ivar designator: B, (required)
    :ivar first_name: C
    :ivar street: D
    :ivar po_box: E
    :ivar postal_code: F
    :ivar city: G
    :ivar state: H
    :ivar country: I
    :ivar phone: J
    :ivar mobile: K
    :ivar fax: L
    :ivar e_mail: M
    """

    class Meta:
        name = "CTR"

    id: None | str = field(
        default=None,
        metadata={
            "name": "A",
            "full_name": "CustomerId",
            "type": "Attribute",
            "required": True,
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(CTR|CTR-)([0-9])+",
        },
    )
    designator: None | str = field(
        default=None,
        metadata={
            "name": "B",
            "full_name": "CustomerDesignator",
            "type": "Attribute",
            "required": True,
            "max_length": 32,
        },
    )
    first_name: None | str = field(
        default=None,
        metadata={
            "name": "C",
            "full_name": "CustomerFirstName",
            "type": "Attribute",
            "max_length": 32,
        },
    )
    street: None | str = field(
        default=None,
        metadata={
            "name": "D",
            "full_name": "CustomerStreet",
            "type": "Attribute",
            "max_length": 32,
        },
    )
    po_box: None | str = field(
        default=None,
        metadata={
            "name": "E",
            "full_name": "CustomerPOBox",
            "type": "Attribute",
            "max_length": 32,
        },
    )
    postal_code: None | str = field(
        default=None,
        metadata={
            "name": "F",
            "full_name": "CustomerPostalCode",
            "type": "Attribute",
            "max_length": 10,
        },
    )
    city: None | str = field(
        default=None,
        metadata={
            "name": "G",
            "full_name": "CustomerCity",
            "type": "Attribute",
            "max_length": 32,
        },
    )
    state: None | str = field(
        default=None,
        metadata={
            "name": "H",
            "full_name": "CustomerState",
            "type": "Attribute",
            "max_length": 32,
        },
    )
    country: None | str = field(
        default=None,
        metadata={
            "name": "I",
            "full_name": "CustomerCountry",
            "type": "Attribute",
            "max_length": 32,
        },
    )
    phone: None | str = field(
        default=None,
        metadata={
            "name": "J",
            "full_name": "CustomerPhone",
            "type": "Attribute",
            "max_length": 20,
        },
    )
    mobile: None | str = field(
        default=None,
        metadata={
            "name": "K",
            "full_name": "CustomerMobile",
            "type": "Attribute",
            "max_length": 20,
        },
    )
    fax: None | str = field(
        default=None,
        metadata={
            "name": "L",
            "full_name": "CustomerFax",
            "type": "Attribute",
            "max_length": 20,
        },
    )
    e_mail: None | str = field(
        default=None,
        metadata={
            "name": "M",
            "full_name": "CustomerEMail",
            "type": "Attribute",
            "max_length": 64,
        },
    )



from isoxml.models.base.v4.allocation_stamp import AllocationStamp, AllocationStampType
from isoxml.models.base.v4.attached_file import AttachedFile, AttachedFilePreserve
from isoxml.models.base.v4.base_station import BaseStation
from isoxml.models.base.v4.coded_comment import CodedComment, CodedCommentScope
from isoxml.models.base.v4.coded_comment_group import CodedCommentGroup
from isoxml.models.base.v4.coded_comment_list_value import CodedCommentListValue
from isoxml.models.base.v4.colour_legend import ColourLegend
from isoxml.models.base.v4.colour_range import ColourRange
from isoxml.models.base.v4.comment_allocation import CommentAllocation
from isoxml.models.base.v4.connection import Connection
from isoxml.models.base.v4.control_assignment import ControlAssignment
from isoxml.models.base.v4.crop_type import CropType
from isoxml.models.base.v4.crop_variety import CropVariety
from isoxml.models.base.v4.cultural_practice import CulturalPractice
from isoxml.models.base.v4.customer import Customer
from isoxml.models.base.v4.data_log_trigger import DataLogTrigger
from isoxml.models.base.v4.data_log_value import DataLogValue
from isoxml.models.base.v4.device import Device
from isoxml.models.base.v4.device_allocation import DeviceAllocation
from isoxml.models.base.v4.device_element import DeviceElement, DeviceElementType
from isoxml.models.base.v4.device_object_reference import DeviceObjectReference
from isoxml.models.base.v4.device_process_data import DeviceProcessData
from isoxml.models.base.v4.device_property import DeviceProperty
from isoxml.models.base.v4.device_value_presentation import DeviceValuePresentation
from isoxml.models.base.v4.external_file_reference import ExternalFileReference, ExternalFileReferenceB
from isoxml.models.base.v4.farm import Farm
from isoxml.models.base.v4.grid import Grid, GridType
from isoxml.models.base.v4.guidance_allocation import GuidanceAllocation
from isoxml.models.base.v4.guidance_group import GuidanceGroup
from isoxml.models.base.v4.guidance_pattern import GuidancePattern, GuidancePatternType, GuidancePatternOptions, \
    GuidancePatternPropagationDirection, GuidancePatternExtension, GuidancePatternGNSSMethod
from isoxml.models.base.v4.guidance_shift import GuidanceShift
from isoxml.models.base.v4.iso11783_link_list import Iso11783LinkList, Iso11783LinkListDataTransferOrigin, \
    Iso11783LinkListVersionMajor, Iso11783LinkListVersionMinor
from isoxml.models.base.v4.iso11783_task_data import Iso11783TaskData, Iso11783TaskDataDataTransferOrigin, \
    Iso11783TaskDataVersionMajor, Iso11783TaskDataVersionMinor
from isoxml.models.base.v4.line_string import LineString, LineStringType
from isoxml.models.base.v4.link import Link
from isoxml.models.base.v4.link_group import LinkGroup, LinkGroupB
from isoxml.models.base.v4.oper_tech_practice import OperTechPractice
from isoxml.models.base.v4.operation_technique import OperationTechnique
from isoxml.models.base.v4.operation_technique_reference import OperationTechniqueReference
from isoxml.models.base.v4.partfield import Partfield
from isoxml.models.base.v4.point import Point, PointType
from isoxml.models.base.v4.polygon import Polygon, PolygonType
from isoxml.models.base.v4.position import Position, PositionStatus
from isoxml.models.base.v4.process_data_variable import ProcessDataVariable
from isoxml.models.base.v4.product import Product, ProductF
from isoxml.models.base.v4.product_allocation import ProductAllocation, ProductAllocationTransferMode
from isoxml.models.base.v4.product_group import ProductGroup, ProductGroupType
from isoxml.models.base.v4.product_relation import ProductRelation
from isoxml.models.base.v4.task import Task, TaskStatus
from isoxml.models.base.v4.task_controller_capabilities import TaskControllerCapabilities, TaskControllerCapabilitiesC
from isoxml.models.base.v4.time import Time, TimeType
from isoxml.models.base.v4.time_log import TimeLog, TimeLogType
from isoxml.models.base.v4.treatment_zone import TreatmentZone
from isoxml.models.base.v4.value_presentation import ValuePresentation
from isoxml.models.base.v4.worker import Worker
from isoxml.models.base.v4.worker_allocation import WorkerAllocation

__all__ = [
    "AttachedFile",
    "AllocationStamp",
    "BaseStation",
    "CommentAllocation",
    "ControlAssignment",
    "CodedCommentGroup",
    "CodedCommentListValue",
    "CodedComment",
    "ColourLegend",
    "Connection",
    "CulturalPractice",
    "ColourRange",
    "CropType",
    "Customer",
    "CropVariety",
    "DeviceAllocation",
    "DeviceElement",
    "DataLogTrigger",
    "DataLogValue",
    "DeviceObjectReference",
    "DeviceProcessData",
    "DeviceProperty",
    "Device",
    "DeviceValuePresentation",
    "Farm",
    "GuidanceAllocation",
    "GuidanceGroup",
    "GuidancePattern",
    "Grid",
    "GuidanceShift",
    "Iso11783TaskData",
    "LineString",
    "OperTechPractice",
    "OperationTechnique",
    "OperationTechniqueReference",
    "ProductAllocation",
    "Product",
    "ProcessDataVariable",
    "Partfield",
    "ProductGroup",
    "Polygon",
    "Point",
    "ProductRelation",
    "Position",
    "TaskControllerCapabilities",
    "Time",
    "TimeLog",
    "Task",
    "TreatmentZone",
    "ValuePresentation",
    "WorkerAllocation",
    "Worker",
    "ExternalFileReference",
    "Iso11783LinkList",
    "LinkGroup",
    "Link",
]



from dataclasses import dataclass, field

from isoxml.models.base.v4.guidance_group import GuidanceGroup
from isoxml.models.base.v4.line_string import LineString
from isoxml.models.base.v4.polygon import Polygon
from isoxml.models.base.v4.point import Point


@dataclass
class Partfield:
    """
    Partfield.

    :ivar polygons: PLN
    :ivar line_strings: LSG
    :ivar points: PNT
    :ivar guidance_groups: GGP
    :ivar id: A, (required)
    :ivar code: B
    :ivar designator: C, (required)
    :ivar area: D, (required)
    :ivar customer_id_ref: E
    :ivar farm_id_ref: F
    :ivar crop_type_id_ref: G
    :ivar crop_variety_id_ref: H
    :ivar field_id_ref: I
    """

    class Meta:
        name = "PFD"

    polygons: list[Polygon] = field(
        default_factory=list,
        metadata={
            "name": "PLN",
            "full_name": "Polygon",
            "type": "Element",
        },
    )
    line_strings: list[LineString] = field(
        default_factory=list,
        metadata={
            "name": "LSG",
            "full_name": "LineString",
            "type": "Element",
        },
    )
    points: list[Point] = field(
        default_factory=list,
        metadata={
            "name": "PNT",
            "full_name": "Point",
            "type": "Element",
        },
    )
    guidance_groups: list[GuidanceGroup] = field(
        default_factory=list,
        metadata={
            "name": "GGP",
            "full_name": "GuidanceGroup",
            "type": "Element",
        },
    )
    id: None | str = field(
        default=None,
        metadata={
            "name": "A",
            "full_name": "PartfieldId",
            "type": "Attribute",
            "required": True,
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(PFD|PFD-)([0-9])+",
        },
    )
    code: None | str = field(
        default=None,
        metadata={
            "name": "B",
            "full_name": "PartfieldCode",
            "type": "Attribute",
            "max_length": 32,
        },
    )
    designator: None | str = field(
        default=None,
        metadata={
            "name": "C",
            "full_name": "PartfieldDesignator",
            "type": "Attribute",
            "required": True,
            "max_length": 32,
        },
    )
    area: None | int = field(
        default=None,
        metadata={
            "name": "D",
            "full_name": "PartfieldArea",
            "type": "Attribute",
            "required": True,
            "min_inclusive": 0,
            "max_inclusive": 4294967294,
        },
    )
    customer_id_ref: None | str = field(
        default=None,
        metadata={
            "name": "E",
            "full_name": "CustomerIdRef",
            "type": "Attribute",
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(CTR|CTR-)([0-9])+",
        },
    )
    farm_id_ref: None | str = field(
        default=None,
        metadata={
            "name": "F",
            "full_name": "FarmIdRef",
            "type": "Attribute",
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(FRM|FRM-)([0-9])+",
        },
    )
    crop_type_id_ref: None | str = field(
        default=None,
        metadata={
            "name": "G",
            "full_name": "CropTypeIdRef",
            "type": "Attribute",
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(CTP|CTP-)([0-9])+",
        },
    )
    crop_variety_id_ref: None | str = field(
        default=None,
        metadata={
            "name": "H",
            "full_name": "CropVarietyIdRef",
            "type": "Attribute",
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(CVT|CVT-)([0-9])+",
        },
    )
    field_id_ref: None | str = field(
        default=None,
        metadata={
            "name": "I",
            "full_name": "FieldIdRef",
            "type": "Attribute",
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(PFD|PFD-)([0-9])+",
        },
    )



from dataclasses import dataclass, field
from enum import Enum

from isoxml.models.base.v4.link import Link


class LinkGroupB(Enum):
    UUIDs = '1'
    ManufacturerProprietary = '2'
    UniqueResolvableURIs = '3'
    InformationalResolvableURIs = '4'


@dataclass
class LinkGroup:
    """
    LinkGroup.

    :ivar links: LNK
    :ivar id: A, (required)
    :ivar type: B, (required)
    :ivar manufacturer_gln: C
    :ivar namespace: D
    :ivar designator: E
    """

    class Meta:
        name = "LGP"

    links: list[Link] = field(
        default_factory=list,
        metadata={
            "name": "LNK",
            "type": "Element",
        },
    )
    id: None | str = field(
        default=None,
        metadata={
            "name": "A",
            "full_name": "LinkGroupId",
            "type": "Attribute",
            "required": True,
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(LGP|LGP-)([0-9])+",
        },
    )
    type: None | LinkGroupB = field(
        default=None,
        metadata={
            "name": "B",
            "full_name": "LinkGroupType",
            "type": "Attribute",
            "required": True,
        },
    )
    manufacturer_gln: None | str = field(
        default=None,
        metadata={
            "name": "C",
            "full_name": "ManufacturerGLN",
            "type": "Attribute",
            "max_length": 64,
        },
    )
    namespace: None | str = field(
        default=None,
        metadata={
            "name": "D",
            "full_name": "LinkGroupNamespace",
            "type": "Attribute",
            "max_length": 255,
        },
    )
    designator: None | str = field(
        default=None,
        metadata={
            "name": "E",
            "full_name": "LinkGroupDesignator",
            "type": "Attribute",
            "max_length": 32,
        },
    )



from decimal import Decimal
from types import ModuleType

import shapely as shp

import isoxml.models.base.v3 as iso3
import isoxml.models.base.v4 as iso4


class _ShapelyConverter:
    def __init__(self, iso_module: ModuleType) -> None:
        self.iso = iso_module

    def _iso_point_from_coords(
            self, east: float, nord: float, pnt_type=None, **kwargs
    ):
        if pnt_type is None:
            pnt_type = self.iso.PointType.Other
        return self.iso.Point(
            type=pnt_type,
            north=round(Decimal(nord), 9),
            east=round(Decimal(east), 9),
            **kwargs
        )

    def to_iso_point(self, point: shp.Point, pnt_type=None, **kwargs):
        if pnt_type is None:
            pnt_type = self.iso.PointType.Other
        return self._iso_point_from_coords(
            east=point.x,
            nord=point.y,
            pnt_type=pnt_type,
            **kwargs
        )

    @staticmethod
    def to_shapely_point(iso_pnt) -> shp.Point:
        return shp.Point(iso_pnt.east, iso_pnt.north)

    def to_iso_line_string(
            self, line: shp.LineString, line_type, **kwargs
    ):
        return self.iso.LineString(
            type=line_type,
            points=[self._iso_point_from_coords(east=x, nord=y) for x, y in line.coords],
            **kwargs
        )

    @staticmethod
    def to_shapely_line_string(iso_line) -> shp.LineString:
        coords = [[iso_pnt.east, iso_pnt.north] for iso_pnt in iso_line.points]
        if iso_line.type == iso3.LineStringType.PolygonExterior or iso_line.type == iso3.LineStringType.PolygonInterior:
            # iso3 LineStrings are implicitly closed if they are of type 1 or 2
            if coords[0] != coords[-1]:
                coords.append(coords[0])
        if len(coords) == 1:  # I could not find any info if this is allowed in v3, but it is in v4 (GPN A+ & Pivot)
            # shapely will accept this, but the geom is not valid
            # will see if that works
            coords.append(coords[0])
        return shp.LineString(coords)

    def to_iso_polygon(
            self, shp_polygon: shp.Polygon, poly_type=None, **kwargs
    ):
        if poly_type is None:
            poly_type = self.iso.PolygonType.Other
        shell = self.to_iso_line_string(shp_polygon.exterior, self.iso.LineStringType.PolygonExterior)
        inner = [self.to_iso_line_string(shp_ls, self.iso.LineStringType.PolygonInterior) for shp_ls in
                 shp_polygon.interiors]
        return self.iso.Polygon(
            type=poly_type,
            line_strings=[shell] + inner,
            **kwargs
        )

    def to_shapely_polygon(self, iso_polygon) -> shp.Polygon:
        holes = []
        shell = None
        for iso_ls in iso_polygon.line_strings:
            shp_ls = self.to_shapely_line_string(iso_ls)
            match iso_ls.type:
                case iso3.LineStringType.PolygonInterior | iso4.LineStringType.PolygonInterior:
                    holes.append(shp_ls)
                case _:
                    # in my understanding, the outer shell may also be implied.
                    # some terminal exports (CNH) contain only one LineString of type flag inside a polygon (bug?)
                    if shell is not None:
                        # v3 allowed multiple outer bound LineStrings in one Polygon
                        # to support this, the association of the holes would have to be done via shapely "contains"
                        # However, I have never seen a polygon that is structured like this
                        raise NotImplementedError("""
                            The conversion of MultiPolygons in the style of v3 is not supported (yet).
                            You can support the development of this library by sending us a 
                            sample ISOXML file that caused this error.
                        """)
                    shell = shp_ls
        if not shell:
            raise ValueError("provides isoxml polygon did not contain a PolygonExterior")
        return shp.Polygon(
            shell=shell,
            holes=holes
        )

    def to_iso_point_list(
            self, multi_point: shp.MultiPoint, pnt_type=None, **kwargs
    ) -> list:
        if pnt_type is None:
            pnt_type = self.iso.PointType.Other
        return [self.to_iso_point(shp_point, pnt_type, **kwargs) for shp_point in multi_point.geoms]

    def to_shapely_multi_point(self, iso_points: list) -> shp.MultiPoint:
        return shp.MultiPoint([self.to_shapely_point(iso_point) for iso_point in iso_points])

    def to_iso_line_string_list(
            self, multi_line: shp.MultiLineString, lsg_type, **kwargs
    ) -> list:
        return [
            self.to_iso_line_string(shp_line, lsg_type, **kwargs)
            for shp_line in multi_line.geoms
        ]

    def to_shapely_multi_line_string(self, iso_lines: list) -> shp.MultiLineString:
        return shp.MultiLineString([
            self.to_shapely_line_string(iso_line)
            for iso_line in iso_lines
        ])

    def to_iso_polygon_list(
            self, multi_polygon: shp.MultiPolygon, pln_type=None, **kwargs
    ) -> list:
        if pln_type is None:
            pln_type = self.iso.PolygonType.Other
        return [
            self.to_iso_polygon(shp_polygon, pln_type, **kwargs)
            for shp_polygon in multi_polygon.geoms
        ]

    def to_shapely_multi_polygon(self, iso_polygons: list) -> shp.MultiPolygon:
        return shp.MultiPolygon([
            self.to_shapely_polygon(iso_polygon)
            for iso_polygon in iso_polygons
        ])

    def to_shapely_geom(self, iso_geometry) -> shp.Geometry:
        match type(iso_geometry):
            case self.iso.Point:
                return self.to_shapely_point(iso_geometry)
            case self.iso.LineString:
                return self.to_shapely_line_string(iso_geometry)
            case self.iso.Polygon:
                return self.to_shapely_polygon(iso_geometry)
            case _:
                raise NotImplementedError(f"Unknown iso geometry {type(iso_geometry)}")


class ShapelyConverterV3(_ShapelyConverter):
    def __init__(self):
        super().__init__(iso3)


class ShapelyConverterV4(_ShapelyConverter):
    def __init__(self):
        super().__init__(iso4)



from dataclasses import dataclass, field
from enum import Enum

from isoxml.models.base.v3.comment_allocation import CommentAllocation
from isoxml.models.base.v3.connection import Connection
from isoxml.models.base.v3.data_log_trigger import DataLogTrigger
from isoxml.models.base.v3.device_allocation import DeviceAllocation
from isoxml.models.base.v3.grid import Grid
from isoxml.models.base.v3.oper_tech_practice import OperTechPractice
from isoxml.models.base.v3.product_allocation import ProductAllocation
from isoxml.models.base.v3.time import Time
from isoxml.models.base.v3.time_log import TimeLog
from isoxml.models.base.v3.treatment_zone import TreatmentZone
from isoxml.models.base.v3.worker_allocation import WorkerAllocation


class TaskStatus(Enum):
    Initial = "1"
    Running = "2"
    Paused = "3"
    Finished = "4"


@dataclass
class Task:
    """
    Task.

    :ivar treatment_zones: TZN
    :ivar times: TIM
    :ivar oper_tech_practices: OTP
    :ivar worker_allocations: WAN
    :ivar device_allocations: DAN
    :ivar connections: CNN
    :ivar product_allocations: PAN
    :ivar data_log_triggers: DLT
    :ivar comment_allocations: CAN
    :ivar time_logs: TLG
    :ivar grids: GRD
    :ivar id: A, (required)
    :ivar designator: B
    :ivar customer_id_ref: C
    :ivar farm_id_ref: D
    :ivar partfield_id_ref: E
    :ivar responsible_worker_id_ref: F
    :ivar status: G, (required)
    :ivar default_treatment_zone_code: H
    :ivar position_lost_treatment_zone_code: I
    :ivar out_of_field_treatment_zone_code: J
    """

    class Meta:
        name = "TSK"

    treatment_zones: list[TreatmentZone] = field(
        default_factory=list,
        metadata={
            "name": "TZN",
            "full_name": "TreatmentZone",
            "type": "Element",
        },
    )
    times: list[Time] = field(
        default_factory=list,
        metadata={
            "name": "TIM",
            "full_name": "Time",
            "type": "Element",
        },
    )
    oper_tech_practices: list[OperTechPractice] = field(
        default_factory=list,
        metadata={
            "name": "OTP",
            "full_name": "OperTechPractice",
            "type": "Element",
        },
    )
    worker_allocations: list[WorkerAllocation] = field(
        default_factory=list,
        metadata={
            "name": "WAN",
            "full_name": "WorkerAllocation",
            "type": "Element",
        },
    )
    device_allocations: list[DeviceAllocation] = field(
        default_factory=list,
        metadata={
            "name": "DAN",
            "full_name": "DeviceAllocation",
            "type": "Element",
        },
    )
    connections: list[Connection] = field(
        default_factory=list,
        metadata={
            "name": "CNN",
            "full_name": "Connection",
            "type": "Element",
        },
    )
    product_allocations: list[ProductAllocation] = field(
        default_factory=list,
        metadata={
            "name": "PAN",
            "full_name": "ProductAllocation",
            "type": "Element",
        },
    )
    data_log_triggers: list[DataLogTrigger] = field(
        default_factory=list,
        metadata={
            "name": "DLT",
            "full_name": "DataLogTrigger",
            "type": "Element",
        },
    )
    comment_allocations: list[CommentAllocation] = field(
        default_factory=list,
        metadata={
            "name": "CAN",
            "full_name": "CommentAllocation",
            "type": "Element",
        },
    )
    time_logs: list[TimeLog] = field(
        default_factory=list,
        metadata={
            "name": "TLG",
            "full_name": "TimeLog",
            "type": "Element",
        },
    )
    grids: list[Grid] = field(
        default_factory=list,
        metadata={
            "name": "GRD",
            "full_name": "Grid",
            "type": "Element",
        },
    )
    id: None | str = field(
        default=None,
        metadata={
            "name": "A",
            "full_name": "TaskId",
            "type": "Attribute",
            "required": True,
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(TSK|TSK-)([0-9])+",
        },
    )
    designator: None | str = field(
        default=None,
        metadata={
            "name": "B",
            "full_name": "TaskDesignator",
            "type": "Attribute",
            "max_length": 32,
        },
    )
    customer_id_ref: None | str = field(
        default=None,
        metadata={
            "name": "C",
            "full_name": "CustomerIdRef",
            "type": "Attribute",
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(CTR|CTR-)([0-9])+",
        },
    )
    farm_id_ref: None | str = field(
        default=None,
        metadata={
            "name": "D",
            "full_name": "FarmIdRef",
            "type": "Attribute",
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(FRM|FRM-)([0-9])+",
        },
    )
    partfield_id_ref: None | str = field(
        default=None,
        metadata={
            "name": "E",
            "full_name": "PartfieldIdRef",
            "type": "Attribute",
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(PFD|PFD-)([0-9])+",
        },
    )
    responsible_worker_id_ref: None | str = field(
        default=None,
        metadata={
            "name": "F",
            "full_name": "ResponsibleWorkerIdRef",
            "type": "Attribute",
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(WKR|WKR-)([0-9])+",
        },
    )
    status: None | TaskStatus = field(
        default=None,
        metadata={
            "name": "G",
            "full_name": "TaskStatus",
            "type": "Attribute",
            "required": True,
        },
    )
    default_treatment_zone_code: None | int = field(
        default=None,
        metadata={
            "name": "H",
            "full_name": "DefaultTreatmentZoneCode",
            "type": "Attribute",
            "min_inclusive": 0,
            "max_inclusive": 254,
        },
    )
    position_lost_treatment_zone_code: None | int = field(
        default=None,
        metadata={
            "name": "I",
            "full_name": "PositionLostTreatmentZoneCode",
            "type": "Attribute",
            "min_inclusive": 0,
            "max_inclusive": 254,
        },
    )
    out_of_field_treatment_zone_code: None | int = field(
        default=None,
        metadata={
            "name": "J",
            "full_name": "OutOfFieldTreatmentZoneCode",
            "type": "Attribute",
            "min_inclusive": 0,
            "max_inclusive": 254,
        },
    )



from dataclasses import dataclass, field

from isoxml.models.base.v4.allocation_stamp import AllocationStamp
from isoxml.models.base.v4.guidance_shift import GuidanceShift


@dataclass
class GuidanceAllocation:
    """
    GuidanceAllocation.

    :ivar allocation_stamps: ASP
    :ivar guidance_shifts: GST
    :ivar guidance_group_id_ref: A, (required)
    """

    class Meta:
        name = "GAN"

    allocation_stamps: list[AllocationStamp] = field(
        default_factory=list,
        metadata={
            "name": "ASP",
            "full_name": "AllocationStamp",
            "type": "Element",
        },
    )
    guidance_shifts: list[GuidanceShift] = field(
        default_factory=list,
        metadata={
            "name": "GST",
            "full_name": "GuidanceShift",
            "type": "Element",
        },
    )
    guidance_group_id_ref: None | str = field(
        default=None,
        metadata={
            "name": "A",
            "full_name": "GuidanceGroupIdRef",
            "type": "Attribute",
            "required": True,
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(GGP|GGP-)([0-9])+",
        },
    )



from dataclasses import dataclass, field
from enum import Enum

from isoxml.models.base.v4.allocation_stamp import AllocationStamp


class ProductAllocationTransferMode(Enum):
    Filling = "1"
    Emptying = "2"
    Remainder = "3"

@dataclass
class ProductAllocation:
    """
    ProductAllocation.

    :ivar asp: ASP
    :ivar product_id_ref: A, (required)
    :ivar quantity_ddi: B
    :ivar quantity_value: C
    :ivar transfer_mode: D
    :ivar device_element_id_ref: E
    :ivar value_presentation_id_ref: F
    :ivar product_sub_type_id_ref: G
    """

    class Meta:
        name = "PAN"

    asp: None | AllocationStamp = field(
        default=None,
        metadata={
            "name": "ASP",
            "type": "Element",
        },
    )
    product_id_ref: None | str = field(
        default=None,
        metadata={
            "name": "A",
            "full_name": "ProductIdRef",
            "type": "Attribute",
            "required": True,
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(PDT|PDT-)([0-9])+",
        },
    )
    quantity_ddi: None | bytes = field(
        default=None,
        metadata={
            "name": "B",
            "full_name": "QuantityDDI",
            "type": "Attribute",
            "length": 2,
            "format": "base16",
        },
    )
    quantity_value: None | int = field(
        default=None,
        metadata={
            "name": "C",
            "full_name": "QuantityValue",
            "type": "Attribute",
            "min_inclusive": 0,
            "max_inclusive": 2147483647,
        },
    )
    transfer_mode: None | ProductAllocationTransferMode = field(
        default=None,
        metadata={
            "name": "D",
            "full_name": "TransferMode",
            "type": "Attribute",
        },
    )
    device_element_id_ref: None | str = field(
        default=None,
        metadata={
            "name": "E",
            "full_name": "DeviceElementIdRef",
            "type": "Attribute",
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(DET|DET-)([0-9])+",
        },
    )
    value_presentation_id_ref: None | str = field(
        default=None,
        metadata={
            "name": "F",
            "full_name": "ValuePresentationIdRef",
            "type": "Attribute",
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(VPN|VPN-)([0-9])+",
        },
    )
    product_sub_type_id_ref: None | str = field(
        default=None,
        metadata={
            "name": "G",
            "full_name": "ProductSubTypeIdRef",
            "type": "Attribute",
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(PDT|PDT-)([0-9])+",
        },
    )






from dataclasses import dataclass, field

from isoxml.models.base.v4.device_element import DeviceElement
from isoxml.models.base.v4.device_process_data import DeviceProcessData
from isoxml.models.base.v4.device_property import DeviceProperty
from isoxml.models.base.v4.device_value_presentation import DeviceValuePresentation


@dataclass
class Device:
    """
    Device.

    :ivar elements: DET
    :ivar properties: DPT
    :ivar process_data: DPD
    :ivar value_presentations: DVP
    :ivar id: A, (required)
    :ivar designator: B
    :ivar software_version: C
    :ivar client_name: D, (required)
    :ivar serial_number: E
    :ivar structure_label: F, (required)
    :ivar localization_label: G, (required)
    """

    class Meta:
        name = "DVC"

    elements: list[DeviceElement] = field(
        default_factory=list,
        metadata={
            "name": "DET",
            "full_name": "DeviceElement",
            "type": "Element",
        },
    )
    properties: list[DeviceProperty] = field(
        default_factory=list,
        metadata={
            "name": "DPT",
            "full_name": "DeviceProperty",
            "type": "Element",
        },
    )
    process_data: list[DeviceProcessData] = field(
        default_factory=list,
        metadata={
            "name": "DPD",
            "full_name": "DeviceProcessData",
            "type": "Element",
        },
    )
    value_presentations: list[DeviceValuePresentation] = field(
        default_factory=list,
        metadata={
            "name": "DVP",
            "full_name": "DeviceValuePresentation",
            "type": "Element",
        },
    )
    id: None | str = field(
        default=None,
        metadata={
            "name": "A",
            "full_name": "DeviceId",
            "type": "Attribute",
            "required": True,
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(DVC|DVC-)([0-9])+",
        },
    )
    designator: None | str = field(
        default=None,
        metadata={
            "name": "B",
            "full_name": "DeviceDesignator",
            "type": "Attribute",
            "max_length": 32,
        },
    )
    software_version: None | str = field(
        default=None,
        metadata={
            "name": "C",
            "full_name": "DeviceSoftwareVersion",
            "type": "Attribute",
            "max_length": 32,
        },
    )
    client_name: None | bytes = field(
        default=None,
        metadata={
            "name": "D",
            "full_name": "ClientNAME",
            "type": "Attribute",
            "required": True,
            "length": 8,
            "format": "base16",
        },
    )
    serial_number: None | str = field(
        default=None,
        metadata={
            "name": "E",
            "full_name": "DeviceSerialNumber",
            "type": "Attribute",
            "max_length": 32,
        },
    )
    structure_label: None | str = field(
        default=None,
        metadata={
            "name": "F",
            "full_name": "DeviceStructureLabel",
            "type": "Attribute",
            "required": True,
            "min_length": 7,
            "max_length": 39,
            "pattern": r"((([0-9]|[a-e]|[A-E])([0-9]|[a-f]|[A-F]))|((F|f)([0-9]|[a-e]|[A-E]))){7}(([0-9]|[a-f]|[A-F])([0-9]|[a-f]|[A-F]))*",
            "format": "base16",
        },
    )
    localization_label: None | str = field(
        default=None,
        metadata={
            "name": "G",
            "full_name": "DeviceLocalizationLabel",
            "type": "Attribute",
            "required": True,
            "length": 7,
            "pattern": r"(F|f){2}((([0-9]|[A-E]|[a-e])([0-9]|[A-F]|[a-f]))|((F|f)([0-9]|[A-E]|[a-e])))*",
            "format": "base16",
        },
    )



from dataclasses import dataclass, field
from enum import Enum


class TimeLogType(Enum):
    BinaryTimelogFileType1 = "1"


@dataclass
class TimeLog:
    """
    TimeLog.

    :ivar filename: A, (required)
    :ivar filelength: B
    :ivar type: C, (required)
    """

    class Meta:
        name = "TLG"

    filename: None | str = field(
        default=None,
        metadata={
            "name": "A",
            "full_name": "Filename",
            "type": "Attribute",
            "required": True,
            "length": 8,
            "pattern": r"TLG[0-9][0-9][0-9][0-9][0-9]",
        },
    )
    filelength: None | int = field(
        default=None,
        metadata={
            "name": "B",
            "full_name": "Filelength",
            "type": "Attribute",
            "min_inclusive": 0,
            "max_inclusive": 4294967294,
        },
    )
    type: None | TimeLogType = field(
        default=None,
        metadata={
            "name": "C",
            "full_name": "TimeLogType",
            "type": "Attribute",
            "required": True,
        },
    )



from dataclasses import dataclass, field
from decimal import Decimal
from enum import Enum


class PointType(Enum):
    Flag = "1"
    Other = "2"


@dataclass
class Point:
    """
    Point.

    :ivar type: A, (required)
    :ivar designator: B
    :ivar north: C, (required)
    :ivar east: D, (required)
    :ivar up: E
    :ivar colour: F
    """

    class Meta:
        name = "PNT"

    type: None | PointType = field(
        default=None,
        metadata={
            "name": "A",
            "full_name": "PointType",
            "type": "Attribute",
            "required": True,
        },
    )
    designator: None | str = field(
        default=None,
        metadata={
            "name": "B",
            "full_name": "PointDesignator",
            "type": "Attribute",
            "max_length": 32,
        },
    )
    north: None | Decimal = field(
        default=None,
        metadata={
            "name": "C",
            "full_name": "PointNorth",
            "type": "Attribute",
            "required": True,
            "min_inclusive": Decimal("-90.0"),
            "max_inclusive": Decimal("90.0"),
        },
    )
    east: None | Decimal = field(
        default=None,
        metadata={
            "name": "D",
            "full_name": "PointEast",
            "type": "Attribute",
            "required": True,
            "min_inclusive": Decimal("-180.0"),
            "max_inclusive": Decimal("180.0"),
        },
    )
    up: None | int = field(
        default=None,
        metadata={
            "name": "E",
            "full_name": "PointUp",
            "type": "Attribute",
            "min_inclusive": -2147483647,
            "max_inclusive": 2147483647,
        },
    )
    colour: None | int = field(
        default=None,
        metadata={
            "name": "F",
            "full_name": "PointColour",
            "type": "Attribute",
            "min_inclusive": 0,
            "max_inclusive": 254,
        },
    )



from dataclasses import dataclass, field

from isoxml.models.base.v3.colour_range import ColourRange


@dataclass
class ColourLegend:
    """
    ColourLegend.

    :ivar colour_ranges: CRG
    :ivar id: A, (required)
    :ivar default_color: B
    """

    class Meta:
        name = "CLD"

    colour_ranges: list[ColourRange] = field(
        default_factory=list,
        metadata={
            "name": "CRG",
            "full_name": "ColourRange",
            "type": "Element",
            "min_occurs": 1,
        },
    )
    id: None | str = field(
        default=None,
        metadata={
            "name": "A",
            "full_name": "ColourLegendId",
            "type": "Attribute",
            "required": True,
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(CLD|CLD-)([0-9])+",
        },
    )
    default_color: None | int = field(
        default=None,
        metadata={
            "name": "B",
            "full_name": "DefaultColor",
            "type": "Attribute",
            "min_inclusive": 0,
            "max_inclusive": 254,
        },
    )



from dataclasses import dataclass, field
from enum import Enum

from isoxml.models.base.v4.coded_comment_list_value import CodedCommentListValue


class CodedCommentScope(Enum):
    Point = '1'
    Global = '2'
    Continuous = '3'


@dataclass
class CodedComment:
    """
    CodedComment.

    :ivar list_values: CCL
    :ivar id: A, (required)
    :ivar designator: B, (required)
    :ivar scope: C, (required)
    :ivar group_id_ref: D
    """

    class Meta:
        name = "CCT"

    list_values: list[CodedCommentListValue] = field(
        default_factory=list,
        metadata={
            "name": "CCL",
            "full_name": "CodedCommentListValue",
            "type": "Element",
        },
    )
    id: None | str = field(
        default=None,
        metadata={
            "name": "A",
            "full_name": "CodedCommentId",
            "type": "Attribute",
            "required": True,
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(CCT|CCT-)([0-9])+",
        },
    )
    designator: None | str = field(
        default=None,
        metadata={
            "name": "B",
            "full_name": "CodedCommentDesignator",
            "type": "Attribute",
            "required": True,
            "max_length": 32,
        },
    )
    scope: None | CodedCommentScope = field(
        default=None,
        metadata={
            "name": "C",
            "full_name": "CodedCommentScope",
            "type": "Attribute",
            "required": True,
        },
    )
    group_id_ref: None | str = field(
        default=None,
        metadata={
            "name": "D",
            "full_name": "CodedCommentGroupIdRef",
            "type": "Attribute",
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(CCG|CCG-)([0-9])+",
        },
    )



import numpy as np

import isoxml.models.base.v3 as iso3
import isoxml.models.base.v4 as iso4
from isoxml.models.ddi_entities import DDEntity


def from_numpy_array(arr: np.ndarray, grid: iso3.Grid | iso4.Grid) -> bytes:
    match grid.type:
        case iso3.GridType.GridType1 | iso4.GridType.GridType1:
            return from_numpy_array_to_type_1(arr, grid)
        case iso3.GridType.GridType2 | iso4.GridType.GridType2:
            return from_numpy_array_to_type_2(arr, grid)
        case _:
            raise NotImplementedError


def to_numpy_array(
        grid_bin: bytes,
        grid: iso3.Grid | iso4.Grid,
        ddi_list: list[DDEntity] | None = None,
        scale=True
) -> np.ndarray:
    """
    Takes a byte encoded ISOXML Grid and converts it back to a numpy array.
    :param grid_bin: the bytes to the ISOXML Grid 1 or 2
    :param grid: the ISOXML Grid element, aka the metadata from the XML file
    :param ddi_list: (optional and only for type 2) a list of DDEntities to performe scaling
    and for korrekt interpretation of multi-value grids.
    :param scale: (only for type 2, default True) whether to scale the values.
    :return: the numpy representation of the grid
    """
    grid_shape = __extract_grid_shape(grid)
    scale_factor = None
    match grid.type:
        case iso3.GridType.GridType1 | iso4.GridType.GridType1:
            return np.frombuffer(grid_bin, dtype=np.uint8).reshape(grid_shape)
        case iso3.GridType.GridType2 | iso4.GridType.GridType2:
            if ddi_list:
                grid_shape = grid_shape + (len(ddi_list),)
                scale_factor = [ddi.bit_resolution for ddi in ddi_list]
            np_arr = np.frombuffer(grid_bin, dtype=np.int32).reshape(grid_shape)
            if scale and scale_factor:
                np_arr = np_arr.astype(np.float32) * scale_factor
            return np_arr
        case _:
            raise NotImplementedError


def __extract_grid_shape(grid):
    grid_shape = (grid.maximum_row, grid.maximum_column)
    return grid_shape


def from_numpy_array_to_type_1(arr: np.ndarray, grid: iso3.Grid | iso4.Grid) -> bytes:
    if arr.dtype != np.uint8:
        raise ValueError("grid type 1 requires uint8")
    grid_shape = __extract_grid_shape(grid)
    if arr.shape != grid_shape:
        raise ValueError("numpy shape dose not match grid shape")
    return arr.tobytes(order="C")


def from_numpy_array_to_type_2(
        arr: np.ndarray,
        grid: iso3.Grid | iso4.Grid,
        ddi_list: list[DDEntity] | None = None,
        scale=True
) -> bytes:
    """
    converts a numpy array into isoxml binary grid type 2 format
    :param arr: the numpy array, dtype must be convertible to int32, dim order must be (row, column, pdv)
    :param grid: the isoxml grid element
    :param ddi_list: a list of DDEntities to performe scaling and for korrekt interpretation of multi-value girds.
    :param scale: whether to scale the values.
    :return: bytes of isoxml binary grid type 2 format
    """
    grid_shape = __extract_grid_shape(grid)
    if ddi_list and len(ddi_list) > 1:
        grid_shape = grid_shape + (len(ddi_list),)
    if arr.shape != grid_shape:
        raise ValueError("numpy shape dose not match grid shape")

    if scale and ddi_list:
        scale_factor = [int(1 / ddi.bit_resolution) for ddi in ddi_list]
        arr = np.round(arr * scale_factor, decimals=0)
        arr_int32 = arr.astype(dtype=np.int32, order='C', casting='unsafe', copy=True)
    else:
        try:
            arr_int32 = arr.astype(dtype=np.int32, order='C', casting='safe', copy=True)
        except (TypeError, ValueError) as e:
            raise ValueError("cant convert given arr to type int16", e)
    return arr_int32.tobytes(order='C')



from dataclasses import dataclass, field

from isoxml.models.base.v3.operation_technique_reference import OperationTechniqueReference


@dataclass
class CulturalPractice:
    """
    CulturalPractice.

    :ivar operation_technique_references: OTR
    :ivar id: A, (required)
    :ivar designator: B, (required)
    """

    class Meta:
        name = "CPC"

    operation_technique_references: list[OperationTechniqueReference] = field(
        default_factory=list,
        metadata={
            "name": "OTR",
            "full_name": "OperationTechniqueReference",
            "type": "Element",
        },
    )
    id: None | str = field(
        default=None,
        metadata={
            "name": "A",
            "full_name": "CulturalPracticeId",
            "type": "Attribute",
            "required": True,
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(CPC|CPC-)([0-9])+",
        },
    )
    designator: None | str = field(
        default=None,
        metadata={
            "name": "B",
            "full_name": "CulturalPracticeDesignator",
            "type": "Attribute",
            "required": True,
            "max_length": 32,
        },
    )



from dataclasses import dataclass, field
from enum import Enum

from isoxml.models.base.v4.link_group import LinkGroup


class Iso11783LinkListDataTransferOrigin(Enum):
    """
    :cvar FMIS: Farm Management Information Systems
    :cvar MICS: mobile implement control system
    """

    FMIS = "1"
    MICS = "2"


class Iso11783LinkListVersionMajor(Enum):
    """
    :cvar VALUE_4: The version of the second edition published as a
        Final Draft International Standard
    """

    VALUE_4 = "4"


class Iso11783LinkListVersionMinor(Enum):
    """
    All minor versions up to 3 of the XML schema for this major version.
    """

    VALUE_0 = "0"
    VALUE_1 = "1"
    VALUE_2 = "2"
    VALUE_3 = "3"


@dataclass
class Iso11783LinkList:
    """
    ISO 11783 Link List File.

    :ivar link_groups: LGP
    :ivar version_major: Version Major, (default: 4)
    :ivar version_minor: Version Minor, (default: 3)
    :ivar management_software_manufacturer: ManagementSoftwareManufacturer, (required)
    :ivar management_software_version: ManagementSoftwareVersion, (required)
    :ivar task_controller_manufacturer: TaskControllerManufacturer
    :ivar task_controller_version: TaskControllerVersion
    :ivar file_version: FileVersion
    :ivar data_transfer_origin: DataTransferOrigin, (required)
    """

    class Meta:
        name = "ISO11783LinkList"

    link_groups: list[LinkGroup] = field(
        default_factory=list,
        metadata={
            "name": "LGP",
            "type": "Element",
        },
    )
    version_major: Iso11783LinkListVersionMajor = field(
        init=False,
        default=Iso11783LinkListVersionMajor.VALUE_4,
        metadata={
            "name": "VersionMajor",
            "type": "Attribute",
            "required": True,
        },
    )
    version_minor: None | Iso11783LinkListVersionMinor = field(
        default=Iso11783LinkListVersionMinor.VALUE_3,
        metadata={
            "name": "VersionMinor",
            "type": "Attribute",
            "required": True,
        },
    )
    management_software_manufacturer: None | str = field(
        default=None,
        metadata={
            "name": "ManagementSoftwareManufacturer",
            "type": "Attribute",
            "required": True,
            "max_length": 32,
        },
    )
    management_software_version: None | str = field(
        default=None,
        metadata={
            "name": "ManagementSoftwareVersion",
            "type": "Attribute",
            "required": True,
            "max_length": 32,
        },
    )
    task_controller_manufacturer: None | str = field(
        default=None,
        metadata={
            "name": "TaskControllerManufacturer",
            "type": "Attribute",
            "max_length": 32,
        },
    )
    task_controller_version: None | str = field(
        default=None,
        metadata={
            "name": "TaskControllerVersion",
            "type": "Attribute",
            "max_length": 32,
        },
    )
    file_version: None | str = field(
        default=None,
        metadata={
            "name": "FileVersion",
            "type": "Attribute",
            "max_length": 32,
        },
    )
    data_transfer_origin: None | Iso11783LinkListDataTransferOrigin = field(
        default=None,
        metadata={
            "name": "DataTransferOrigin",
            "type": "Attribute",
            "required": True,
        },
    )



from dataclasses import dataclass, field


@dataclass
class CropVariety:
    """
    CropVariety.

    :ivar id: A, (required)
    :ivar designator: B, (required)
    :ivar product_id_ref: C
    """

    class Meta:
        name = "CVT"

    id: None | str = field(
        default=None,
        metadata={
            "name": "A",
            "full_name": "CropVarietyId",
            "type": "Attribute",
            "required": True,
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(CVT|CVT-)([0-9])+",
        },
    )
    designator: None | str = field(
        default=None,
        metadata={
            "name": "B",
            "full_name": "CropVarietyDesignator",
            "type": "Attribute",
            "required": True,
            "max_length": 32,
        },
    )
    product_id_ref: None | str = field(
        default=None,
        metadata={
            "name": "C",
            "full_name": "ProductIdRef",
            "type": "Attribute",
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(PDT|PDT-)([0-9])+",
        },
    )



from dataclasses import dataclass, field


@dataclass
class DataLogTrigger:
    """
    DataLogTrigger.

    :ivar data_log_ddi: A, (required)
    :ivar data_log_method: B, (required)
    :ivar data_log_distance_interval: C
    :ivar data_log_time_interval: D
    :ivar data_log_threshold_minimum: E
    :ivar data_log_threshold_maximum: F
    :ivar data_log_threshold_change: G
    :ivar device_element_id_ref: H
    :ivar value_presentation_id_ref: I
    :ivar data_log_pgn: J
    :ivar data_log_pgn_start_bit: K
    :ivar data_log_pgn_stop_bit: L
    """

    class Meta:
        name = "DLT"

    data_log_ddi: None | bytes = field(
        default=None,
        metadata={
            "name": "A",
            "full_name": "DataLogDDI",
            "type": "Attribute",
            "required": True,
            "length": 2,
            "format": "base16",
        },
    )
    data_log_method: None | int = field(
        default=None,
        metadata={
            "name": "B",
            "full_name": "DataLogMethod",
            "type": "Attribute",
            "required": True,
            "min_inclusive": 1,
            "max_inclusive": 31,
        },
    )
    data_log_distance_interval: None | int = field(
        default=None,
        metadata={
            "name": "C",
            "full_name": "DataLogDistanceInterval",
            "type": "Attribute",
            "min_inclusive": 0,
            "max_inclusive": 1000000,
        },
    )
    data_log_time_interval: None | int = field(
        default=None,
        metadata={
            "name": "D",
            "full_name": "DataLogTimeInterval",
            "type": "Attribute",
            "min_inclusive": 0,
            "max_inclusive": 60000,
        },
    )
    data_log_threshold_minimum: None | int = field(
        default=None,
        metadata={
            "name": "E",
            "full_name": "DataLogThresholdMinimum",
            "type": "Attribute",
            "min_inclusive": -2147483647,
            "max_inclusive": 2147483647,
        },
    )
    data_log_threshold_maximum: None | int = field(
        default=None,
        metadata={
            "name": "F",
            "full_name": "DataLogThresholdMaximum",
            "type": "Attribute",
            "min_inclusive": -2147483647,
            "max_inclusive": 2147483647,
        },
    )
    data_log_threshold_change: None | int = field(
        default=None,
        metadata={
            "name": "G",
            "full_name": "DataLogThresholdChange",
            "type": "Attribute",
            "min_inclusive": -2147483647,
            "max_inclusive": 2147483647,
        },
    )
    device_element_id_ref: None | str = field(
        default=None,
        metadata={
            "name": "H",
            "full_name": "DeviceElementIdRef",
            "type": "Attribute",
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(DET|DET-)([0-9])+",
        },
    )
    value_presentation_id_ref: None | str = field(
        default=None,
        metadata={
            "name": "I",
            "full_name": "ValuePresentationIdRef",
            "type": "Attribute",
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(VPN|VPN-)([0-9])+",
        },
    )
    data_log_pgn: None | int = field(
        default=None,
        metadata={
            "name": "J",
            "full_name": "DataLogPGN",
            "type": "Attribute",
            "min_inclusive": 0,
            "max_inclusive": 262143,
        },
    )
    data_log_pgn_start_bit: None | int = field(
        default=None,
        metadata={
            "name": "K",
            "full_name": "DataLogPGNStartBit",
            "type": "Attribute",
            "min_inclusive": 0,
            "max_inclusive": 63,
        },
    )
    data_log_pgn_stop_bit: None | int = field(
        default=None,
        metadata={
            "name": "L",
            "full_name": "DataLogPGNStopBit",
            "type": "Attribute",
            "min_inclusive": 0,
            "max_inclusive": 63,
        },
    )



from dataclasses import dataclass, field


@dataclass
class Farm:
    """
    Farm.

    :ivar id: A, (required)
    :ivar designator: B, (required)
    :ivar street: C
    :ivar po_box: D
    :ivar postal_code: E
    :ivar city: F
    :ivar state: G
    :ivar country: H
    :ivar customer_id_ref: I
    """

    class Meta:
        name = "FRM"

    id: None | str = field(
        default=None,
        metadata={
            "name": "A",
            "full_name": "FarmId",
            "type": "Attribute",
            "required": True,
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(FRM|FRM-)([0-9])+",
        },
    )
    designator: None | str = field(
        default=None,
        metadata={
            "name": "B",
            "full_name": "FarmDesignator",
            "type": "Attribute",
            "required": True,
            "max_length": 32,
        },
    )
    street: None | str = field(
        default=None,
        metadata={
            "name": "C",
            "full_name": "FarmStreet",
            "type": "Attribute",
            "max_length": 32,
        },
    )
    po_box: None | str = field(
        default=None,
        metadata={
            "name": "D",
            "full_name": "FarmPOBox",
            "type": "Attribute",
            "max_length": 32,
        },
    )
    postal_code: None | str = field(
        default=None,
        metadata={
            "name": "E",
            "full_name": "FarmPostalCode",
            "type": "Attribute",
            "max_length": 10,
        },
    )
    city: None | str = field(
        default=None,
        metadata={
            "name": "F",
            "full_name": "FarmCity",
            "type": "Attribute",
            "max_length": 32,
        },
    )
    state: None | str = field(
        default=None,
        metadata={
            "name": "G",
            "full_name": "FarmState",
            "type": "Attribute",
            "max_length": 32,
        },
    )
    country: None | str = field(
        default=None,
        metadata={
            "name": "H",
            "full_name": "FarmCountry",
            "type": "Attribute",
            "max_length": 32,
        },
    )
    customer_id_ref: None | str = field(
        default=None,
        metadata={
            "name": "I",
            "full_name": "CustomerIdRef",
            "type": "Attribute",
            "min_length": 4,
            "max_length": 14,
            "pattern": r"(CTR|CTR-)([0-9])+",
        },
    )
